package checkers.util;

import checkers.types.*;

import java.util.*;

import javax.lang.model.element.*;
import javax.lang.model.type.*;
import javax.lang.model.util.*;

/**
 * A utility class for working with annotations.
 */
public class AnnotationUtils {

    /**
     * A generic base class for parsers of annotation values.
     */
    private abstract static class AbstractAnnotationValueParser<A>
        extends SimpleAnnotationValueVisitor6<Void, Boolean> {

        /**
         * @return the value of an annotation field
         */
        public abstract A getValue();

        public Void visitArray(List<? extends AnnotationValue> vals, Boolean p) {
            if (p != null && p)
                return null;

            for (AnnotationValue a : vals)
                visit(a, Boolean.TRUE);
            
            return null;
        }

    }

    private static class EnumConstantArrayValueParser<R extends Enum<R>>
        extends AbstractAnnotationValueParser<Set<R>> {

        private Set<R> values = new HashSet<R>();
        private Class<R> enumType;

        public EnumConstantArrayValueParser(Class<R> enumType) {
            this.enumType = enumType;
        }
        
        @Override
        public Void visitEnumConstant(VariableElement c, Boolean p) {
            if (p == null || !p)
                return null;

            values.add(Enum.<R>valueOf(enumType, c.getSimpleName().toString()));
            return null;
        }

        public Set<R> getValue() {
            return Collections.unmodifiableSet(values);
        }
    }
    
    /**
     * A utility class for parsing a String-valued annotation.
     */
    private static class StringValueParser
        extends AbstractAnnotationValueParser<String> {

        private String value = null;

        @Override
        public Void visitString(String s, Boolean p) {
            value = s;
            return null;
        }

        public String getValue() {
            return value;
        }
    }

    /**
     * A utility class for parsing a String[]-valued annotation.
     */
    private static class StringArrayValueParser 
        extends AbstractAnnotationValueParser<Set<String>> {

        private final Set<String> values = new HashSet<String>();

        @Override
        public Void visitString(String s, Boolean p) {
            if (p == null || !p)
                return null;

            values.add(s);

            return null;
        }

        public Set<String> getValue() {
            return Collections.unmodifiableSet(values);
        }
    }

    /**
     * A generalized method for obtaining annotation values using any parser
     * that operates on any field.
     *
     * @param <R> the type of value to parse
     * @param parser the annotation value parser
     * @param ad the annotation for which a field will be parsed
     * @param fieldName the name of the annotation field for which a value will be returned
     * @return the value of {@code fieldName} in {@code ad} as determined by
     *         {@code parser}, with type {@code R}
     */
    private static <R> R parseAnnotationValue(AbstractAnnotationValueParser<R> parser,
            AnnotationData ad, String fieldName) {

        Map<? extends ExecutableElement, ? extends AnnotationValue> values =
            ad.getValues();

        for (Map.Entry<? extends ExecutableElement, ? extends
                AnnotationValue> entry : values.entrySet()) {

            ExecutableElement name = entry.getKey();
            AnnotationValue value = entry.getValue();

            {
                Name eltName = name.getSimpleName();
                if (!(fieldName.equals(eltName.toString())))
                    continue;
            }

            parser.visit(value);
            return parser.getValue();
        }
        
        return null;
    }
    
    /**
     * @param <R> the enum type
     * @param ad the annotation for which a value will be parsed
     * @param field the name of the field to parse
     * @param enumType the type of the enum
     * @return the enum constant values of the given field
     */
    public static <R extends Enum<R>> Set<R> parseEnumConstantArrayValue(AnnotationData ad, String field, Class<R> enumType) {
        return parseAnnotationValue(new EnumConstantArrayValueParser<R>(enumType), ad, field);
    }
    
    /**
     * @param ad the annotation for which a value will be parsed
     * @param field the name of the field to parse
     * @return the String value of the given field
     */
    public static String parseStringValue(AnnotationData ad, String field) {
        return parseAnnotationValue(new StringValueParser(), ad, field);
    }

    /**
     * @param ad the annotation for which a value will be parsed
     * @param field the name of the field to parse
     * @return the String values of the given field
     */
    public static Set<String> parseStringArrayValue(AnnotationData ad, String field) {
        return parseAnnotationValue(new StringArrayValueParser(), ad, field);
    }
    
    /**
     * For an annotation with a "value" field of type String[], determines the
     * String values as a {@link Set}.
     *
     * @param ad the annotation to parse
     * @return a {@link Set} of {@link String}s for the annotation's String[]
     *         values, or an empty {@link Set} if the annotation doesn't have a
     *         String[] value field, or null if the annotation doesn't have a
     *         value field at all
     *
     * @deprecated use {@link parseStringArrayValue(AnnotationData, String)}
     *             instead
     */
    @Deprecated
    public static Set<String> parseStringArrayValue(AnnotationData ad) {
        return parseStringArrayValue(ad, "value");
    }

    /**
     * @return the fully-qualified name of an annotation as a String
     */
    public static final String annotationName(AnnotationData annotation) {
        TypeMirror annoType = annotation.getType();
        if (annoType.getKind() != TypeKind.DECLARED)
            return null;

        DeclaredType dt = (DeclaredType)annoType;
        Element elt = dt.asElement();
        if (elt instanceof TypeElement)
            return ((TypeElement)elt).getQualifiedName().toString();
        return null;
    }
}
