package tests;

import javax.tools.*;
import java.io.*;
import java.util.*;

import static org.junit.Assert.*;

import checkers.quals.*;

/**
 * Abstract class for testing a checker framework.
 */
abstract public class CheckerTest {

    /**
     * Set this variable to the checker name.
     */
    protected String checkerName;

    /**
     * Set this variable to the subdirectory.
     */
    protected String checkerDir;

    /**
     * Compiles and returns a TestRun.
     */
    protected TestRun getTest(String... files) {

        List<String> fileStrings = new LinkedList<String>();
        for (String s : files)
            fileStrings.add(checkerDir + "/" + s);

        JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();
        StandardJavaFileManager fileManager
            = compiler.getStandardFileManager(null, null, null);
        Iterable<? extends JavaFileObject> tests =
            fileManager.getJavaFileObjectsFromStrings(fileStrings);

        return
            TestInput.compileAndCheck(tests, checkerName);
    }

    /**
     * Asserts that the test compilation completed without failures or
     * exceptions.
     *
     * @param run the test run to check
     */
    protected void assertSuccess(@ReadOnly TestRun run) {
        assertTrue("failures:\n" + run.getDiagnostics(), run.getResult());
    }

    /**
     * Asserts that the test compilation did not complete successfully.
     *
     * @param run the test run to check
     */
    protected void assertFailure(@ReadOnly TestRun run) {
        assertFalse("the test run was successful", run.getResult());
    }

    /**
     * Compares the result of the compiler against a list of errors in a file.
     * If the file is not found or cannot be read, the assertion fails.
     *
     * @param list the list of diagnostics from the compiler
     * @param diagnosticFile a file containing a list of expected errors, one
     *        per line
     */
    protected void assertDiagnostics(@ReadOnly List<@ReadOnly Diagnostic<@ReadOnly ? extends JavaFileObject>> list,
            @ReadOnly String diagnosticFile) {

        try {
            String filePath = checkerDir + File.separator + diagnosticFile;
            BufferedReader reader = new BufferedReader(new FileReader(filePath));
            ArrayList<String> lines = new ArrayList<String>();
            String line;
            while ((line = reader.readLine()) != null)
                lines.add(line.trim());
            assertDiagnostics(list, lines.toArray(new String[lines.size()]));
        } catch (IOException e) {
            fail(e.getMessage());
        }
    }

    /**
     * Compares the result of the compiler against an array of Strings.
     */
    protected void assertDiagnostics(@ReadOnly List<@ReadOnly Diagnostic<@ReadOnly ? extends JavaFileObject>> list, @ReadOnly String[@ReadOnly] diagnostics) {
        String cs = (checkerDir == "" ? "" : checkerDir + "/");

        List<String> expectedList = new LinkedList<String>();
        for (@ReadOnly String sd : diagnostics) expectedList.add(cs + sd);

        List<String> resultsList = new LinkedList<String>();
        for (@ReadOnly Diagnostic<@ReadOnly ? extends JavaFileObject> d : list)
            resultsList.add(d.toString());

        List<String> foundList = new LinkedList<String>();
        foundList.addAll(resultsList);
        foundList.retainAll(expectedList);

        String failMessage = foundList.size() + " out of "
            + expectedList.size() + " expected diagnostics "
            + (foundList.size() == 1 ? "was" : "were") +" found.\n";
        boolean failed = false;

        List<String> notFoundList = new LinkedList<String>();
        notFoundList.addAll(expectedList);
        notFoundList.removeAll(resultsList);

        if (!notFoundList.isEmpty()) {
            failed = true;

            String message = notFoundList.size() == 1?
                "1 expected diagnostic was not found:\n" :
                notFoundList.size() + " expected diagnostics were not found:\n";
            failMessage += "\n" + message;

            for (String a : notFoundList)
                failMessage += a + "\n";

        }


        List<String> unexpectedList = new LinkedList<String>();
        unexpectedList.addAll(resultsList);
        unexpectedList.removeAll(expectedList);

        if (!unexpectedList.isEmpty()) {
            failed = true;

            String message = unexpectedList.size() == 1?
                "1 unexpected diagnostic was found:\n" :
                unexpectedList.size() + " unexpected diagnostics were found:\n";
            failMessage += "\n" + message;

            for (String a : unexpectedList)
                failMessage += a + "\n";
        }

        if (failed)
            fail(failMessage);

    }

}
